@@ -198,7 +198,7 @@ impl LangTester {
     }
 
     /// Specify a function which takes a `Path` to a test file and returns a vector containing 1 or
-    /// more `(<name>, <[Command](https://doc.rust-lang.org/std/process/struct.Command.html)>)
+    /// more (`name`, <[`Command`](https://doc.rust-lang.org/std/process/struct.Command.html)>)
     /// pairs. The commands will be executed in order on the test file: for each executed command,
     /// test commands starting with `<name>` will be checked. For example, if your pipeline
     /// requires separate compilation and linking, you might specify something along the lines of
@@ -486,6 +486,7 @@ pub(crate) struct TestCmd<'a> {
     /// A list of custom command line arguments which should be passed when
     /// executing the test command.
     pub args: Vec<String>,
+    pub env: HashMap<String, String>,
 }
 
 impl<'a> TestCmd<'a> {
@@ -496,6 +497,7 @@ impl<'a> TestCmd<'a> {
             stderr: vec!["..."],
             stdout: vec!["..."],
             args: Vec::new(),
+            env: HashMap::new(),
         }
     }
 }
@@ -616,7 +618,6 @@ fn test_file(
                         stdout_match: None,
                     };
                     failures.lock().unwrap().push((test_fname, failure));
-                    return;
                 }
             };
         });
@@ -685,6 +686,7 @@ fn run_tests(
         let default_test = TestCmd::default();
         let test = tests.get(&cmd_name).unwrap_or(&default_test);
         cmd.args(&test.args);
+        cmd.envs(&test.env);
         let (status, stdin_remaining, stderr, stdout) =
             run_cmd(inner.clone(), &test_fname, cmd, &test);
 
@@ -810,14 +812,13 @@ fn run_tests(
                 .set_color(ColorSpec::new().set_fg(Some(Color::Red)))
                 .ok();
             handle.write_all(b"FAILED").ok();
-            handle.reset().ok();
         } else {
             handle
                 .set_color(ColorSpec::new().set_fg(Some(Color::Green)))
                 .ok();
             handle.write_all(b"ok").ok();
-            handle.reset().ok();
         }
+        handle.reset().ok();
     }
 
     false
@@ -841,9 +842,9 @@ fn run_cmd(
         .spawn()
         .unwrap_or_else(|_| fatal(&format!("Couldn't run command {:?}.", cmd)));
 
-    let stdin = child.stdin.as_mut().unwrap();
-    let stderr = child.stderr.as_mut().unwrap();
-    let stdout = child.stdout.as_mut().unwrap();
+    let mut stdin = child.stdin.take().unwrap();
+    let mut stderr = child.stderr.take().unwrap();
+    let mut stdout = child.stdout.take().unwrap();
 
     let stdin_fd = stdin.as_raw_fd();
     let stderr_fd = stderr.as_raw_fd();
@@ -907,9 +908,13 @@ fn run_cmd(
                 }
                 if stdin_off == stdin_str.len() {
                     stdin_finished = true;
-                    // This is a bit icky, because the `stdin` variable will later close stdin when the
-                    // variable is dropped. However, some programs expect stdin to be closed before
-                    // they'll continue, so this is the least worst option.
+                    // If we don't close the child's stdin now, the child process might hang,
+                    // waiting forever for input that will never come. However, if we're not
+                    // careful, we'll close the underlying file handle twice: once here and once
+                    // when Rust drops the `stdin` variable. Since file handles can be reused that
+                    // means we could close a file handle that's now being used elsewhere. Thus
+                    // because we close the child's stdin here, we need to make sure we `forget`
+                    // the `stdin` variable at every return point from this function.
                     unsafe {
                         close(stdin_fd);
                     }
@@ -1009,6 +1014,9 @@ fn run_cmd(
     };
 
     let stdin_remaining = if stdin_finished {
+        // We closed the child's stdin earlier, so we must make sure not to close it again, since
+        // the file handle is stale and may have been reused in the interim.
+        std::mem::forget(stdin);
         0
     } else {
         let stdin_str = test.stdin.as_ref().unwrap();
